// Copyright 2015-2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include "xud_device.h"

/* Endpoint type tables - informs XUD what the transfer types for each Endpoint in use and also
 * if the endpoint wishes to be informed of USB bus state
 */
#ifdef BENCHMARK
#define EP_COUNT_OUT   (1)
#define EP_COUNT_IN    (2)

/* Endpoint types for BENCHMARK config */
XUD_EpType epTypeTableOut[EP_COUNT_OUT] = {XUD_EPTYPE_CTL | XUD_STATUS_ENABLE};
XUD_EpType epTypeTableIn[EP_COUNT_IN] =   {XUD_EPTYPE_CTL | XUD_STATUS_ENABLE, XUD_EPTYPE_BUL};
#else
/* Endpoint counts for BULKTEST config */
#define EP_COUNT_OUT   (2)
#define EP_COUNT_IN    (2)

/* Endpoint types for BULKTEST config */
XUD_EpType epTypeTableOut[EP_COUNT_OUT] = {XUD_EPTYPE_CTL | XUD_STATUS_ENABLE, XUD_EPTYPE_BUL | XUD_STATUS_ENABLE};
XUD_EpType epTypeTableIn[EP_COUNT_IN] =   {XUD_EPTYPE_CTL | XUD_STATUS_ENABLE, XUD_EPTYPE_BUL | XUD_STATUS_ENABLE};
#endif

#define BUFFER_SIZE_WORDS (128)
#define BUFFER_SIZE_CHARS (BUFFER_SIZE_WORDS * sizeof(char))

#ifdef BENCHMARK
/* An optimised endpoint for the read benchmark test */
void bulk_endpoint(chanend chan_ep_to_host)
{
    char host_transfer_buf[BUFFER_SIZE_WORDS*4];
    unsigned host_transfer_length = 512;

    XUD_ep ep_to_host = XUD_InitEp(chan_ep_to_host);

    while(1)
    {
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
        XUD_SetBuffer(ep_to_host, host_transfer_buf, host_transfer_length);
    }
}
#else
/* A basic endpoint function that receives 512-byte packets of data, processes
 * them and sends them back to the host. If at any point an error is detected
 * (return value < 0) then the process needs to be started again so that
 * both host and device stay in .
 */
void bulk_endpoint(chanend chan_ep_from_host, chanend chan_ep_to_host)
{
    int host_transfer_buf[BUFFER_SIZE_WORDS];
    unsigned host_transfer_length = 0;
    XUD_Result_t result;

    XUD_ep ep_from_host = XUD_InitEp(chan_ep_from_host);
    XUD_ep ep_to_host = XUD_InitEp(chan_ep_to_host);

    while(1)
    {
        /* Receive a buffer (512-bytes) of data from the host */
        if((result = XUD_GetBuffer(ep_from_host, (host_transfer_buf, char[BUFFER_SIZE_CHARS]), host_transfer_length)) == XUD_RES_RST)
        {
            XUD_ResetEndpoint(ep_from_host, ep_to_host);
            continue;
        }

        /* Perform basic processing (increment data values) */
        for (int i = 0; i < host_transfer_length/4; i++)
            host_transfer_buf[i]++;

        /* Send the modified buffer back to the host */
        if((result = XUD_SetBuffer(ep_to_host, (host_transfer_buf, char[BUFFER_SIZE_CHARS]), host_transfer_length)) == XUD_RES_RST)
        {
            XUD_ResetEndpoint(ep_from_host, ep_to_host);
        }
    }
}
#endif

/* Function in endpoint0.xc */
void Endpoint0(chanend c_ep0_out, chanend c_ep0_in);

/* The main function runs three tasks: the XUD manager, Endpoint 0, and bulk
 * endpoint. An array of channels is used for both IN and OUT endpoints,
 * endpoint zero requires both, bulk_endpoint requires an IN and an OUT endpoint
 * to receive and send a data buffer to the host */
int main()
{
    /* Declare channels for each endpoint */
    chan c_ep_out[EP_COUNT_OUT], c_ep_in[EP_COUNT_IN];

    par
    {
        on USB_TILE: XUD_Main(c_ep_out, EP_COUNT_OUT, c_ep_in, EP_COUNT_IN,
                      null, epTypeTableOut, epTypeTableIn,
                      XUD_SPEED_HS, XUD_PWR_BUS);

        on USB_TILE: Endpoint0(c_ep_out[0], c_ep_in[0]);

#ifdef BENCHMARK
        on USB_TILE: bulk_endpoint(c_ep_in[1]);
#else
        on USB_TILE: bulk_endpoint(c_ep_out[1], c_ep_in[1]);
#endif
    }

    return 0;
}
